from typing import List, Optional, Type

from ...api import NounPhrase
from ...reflection import BookDescriptor
from ..book_config_descriptor import BookConfigDescriptor


class BookFactory:
    @classmethod
    def create(
        cls,
        t: Type,
        identifier: Optional[str] = None,
        name: Optional[str] = None,
        noun_phrase_str: Optional[str] = None,
        author: Optional[str] = None,
        short_description: Optional[str] = None,
        long_description: Optional[str] = None,
        icon: Optional[bytes] = None,
        icon_path: Optional[str] = None,
        configuration: Optional[List[BookConfigDescriptor]] = None,
        tags: Optional[List[str]] = None,
    ) -> BookDescriptor:
        if name is None:
            name = t.__name__

        if identifier is None:
            identifier = name.strip().replace(" ", "").lower()

        if noun_phrase_str is None:
            noun_phrase = NounPhrase.from_pascal_case(name)
        else:
            noun_phrase = NounPhrase.from_head(noun_phrase_str)

        return BookDescriptor(
            cls=t,
            identifier=identifier,
            name=name,
            noun_phrase=noun_phrase,
            author=author,
            short_description=short_description,
            long_description=long_description,
            icon=icon,
            icon_path=icon_path,
            configuration=configuration,
            tags=tags,
        )



from typing import Any, List, Optional

from ...api import NounPhrase
from ...docstring import Docstring
from ..book_config_descriptor import BookConfigDescriptor
from .concept import ConceptFactory


class BookConfigFactory:
    @classmethod
    def create(
        cls, property_name: str, noun_phrases: List[NounPhrase], return_annotation: type, docstring: Optional[Docstring], default_value: Optional[Any] = None
    ) -> BookConfigDescriptor:
        concept = ConceptFactory.from_noun_phrases_and_annotation(
            noun_phrases, return_annotation, description=docstring.short_description if docstring else None, default_value=default_value
        )

        return BookConfigDescriptor(python_property_name=property_name, concept=concept)



from dataclasses import dataclass

from ..api.noun_phrase import NounPhrases
from .types import ConceptType


@dataclass
class QuestionDescriptor:
    noun_phrases: NounPhrases
    type: ConceptType

    def __hash__(self) -> int:
        return hash(self.noun_phrases)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, QuestionDescriptor) and hash(self) == hash(other)



import inspect


def search_hint(text: str):
    """Decorator to register a search hint string for a procedure.

    Can be stacked multiple times on a function decorated with @procedure.
    Stores the raw strings without parsing.
    """
    if not isinstance(text, str):
        raise TypeError("search_hint must be provided a string")

    def decorator(fn):
        if not inspect.isfunction(fn):
            raise TypeError("The search_hint decorator can only be applied to functions.")
        existing = getattr(fn, "__search_hints__", None)
        if existing is None:
            setattr(fn, "__search_hints__", [text])
        else:
            # Insert at the front to preserve source order when stacking decorators
            existing.insert(0, text)
        return fn

    return decorator



# type: ignore [reportReturnType]
from dataclasses import dataclass
from typing import IO, List, Optional, Union

from kognitos.bdk.api import FilterExpression
from kognitos.bdk.decorators import blueprint, blueprint_procedure, concept


@dataclass
@concept(is_a="file reference")
class FileReference:
    """
    Contains all information required to identify a file within a storage system
    """


@dataclass
@concept(is_a="folder reference")
class FolderReference:
    """
    Contains all information required to identify a folder within a storage system
    """


@blueprint
class FileStorageBlueprint:
    """
    A blueprint for files.
    """

    @blueprint_procedure("to get a (root folder)")
    def get_root_folder(self) -> FolderReference:
        """
        Gets a reference to the root folder.

        Returns:
            a folder reference
        """

    @blueprint_procedure("to get a (folder) at a path")
    def get_folder(self, path: str) -> FolderReference:
        """
        Gets a reference to a folder.

        Input Concepts:
            the path: The path to the folder

        Returns:
            a folder reference
        """

    @blueprint_procedure("to get some (folder's items)")
    def list_items(self, folder: FolderReference, filter_expression: Optional[FilterExpression] = None) -> List[Union[FileReference, FolderReference]]:
        """
        Lists items from a folder reference

        Input Concepts:
            the folder: The folder reference from which to list the items
            the filter expression: A filter expression to filter the items in the folder

        Returns:
            a list of Items (file or folder) containing the items in the folder
        """

    @blueprint_procedure("to rename an item to a name")
    def rename_item(self, item: Union[FileReference, FolderReference], name: str, conflict_behavior: Optional[str] = "fail") -> None:
        """
        Rename an item (file or folder) to a given name

        Input Concepts:
            the item: The item (file or folder) to rename
            the name: The new name of the item
            the conflict behavior: The behavior to use on conflict scenarios. It should be one of the following values: ('fail', 'replace', 'rename')
        """

    @blueprint_procedure("to delete an item")
    def delete_an_item(self, item: Union[FileReference, FolderReference]) -> None:
        """
        Delete an item (file or folder)

        Input Concepts:
            the item: The item (file or folder) to delete
        """

    @blueprint_procedure("to download a file")
    def download_item(self, file: FileReference) -> IO:
        """
        Download a file

        Input Concepts:
            the file: The file reference to the file to download

        Returns:
            the file as an IO object
        """

    @blueprint_procedure("to create a (folder) in another folder")
    def create_folder(self, another_folder: FolderReference, folder_name: str, conflict_behavior: Optional[str] = "fail") -> FolderReference:
        """
        Create a (folder) in another folder

        Input Concepts:
            the another folder: The folder to create the (folder) in
            the folder name: The name of the (folder) to create
            the conflict behavior: The behavior to use on conflict scenarios. It should be one of the following values: ('fail', 'replace', 'rename')

        Returns:
            a folder reference to the created folder
        """

    @blueprint_procedure("to upload a (file) to a folder")
    def upload_file_to_folder(self, file: IO, folder: FolderReference, file_name: str, conflict_behavior: Optional[str] = "fail") -> FileReference:
        """
        Upload a file to a folder

        Input Concepts:
            the file: The file to upload
            the folder: The folder to upload the file to
            the file name: The name of the file to upload
            the conflict behavior: The behavior to use on conflict scenarios. It should be one of the following values: ('fail', 'replace', 'rename')

        Returns:
            a file reference to the uploaded file
        """

    @blueprint_procedure("to copy an item to a folder")
    def copy_item(self, item: Union[FileReference, FolderReference], folder: FolderReference, conflict_behavior: Optional[str] = "fail") -> None:
        """
        Copy an item to a folder

        Input Concepts:
            the item: The item (file or folder) to copy
            the folder: The folder to copy the item to
            the conflict behavior: The behavior to use on conflict scenarios. It should be one of the following values: ('fail', 'replace', 'rename')
        """

    @blueprint_procedure("to move an item to a folder")
    def move_item(self, item: Union[FileReference, FolderReference], folder: FolderReference, conflict_behavior: Optional[str] = "fail") -> None:
        """
        Move an item to a folder

        Input Concepts:
            the item: The item (file or folder) to move
            the folder: The folder to move the item to
            the conflict behavior: The behavior to use on conflict scenarios. It should be one of the following values: ('fail', 'replace', 'rename')
        """



from dataclasses import dataclass, field
from typing import Any, List, Optional

from ..api.noun_phrase import NounPhrase
from .types import ConceptOptionalType, ConceptType


@dataclass
class ConceptDescriptor:
    noun_phrases: List[NounPhrase]
    _type: ConceptType = field(init=False)
    default_value: Any
    description: Optional[str] = None

    def __init__(self, noun_phrases: List[NounPhrase], type: ConceptType, default_value: Any, description: Optional[str] = None):  # pylint: disable=redefined-builtin
        self.noun_phrases = noun_phrases
        self._type = type
        self.default_value = default_value
        self.description = description

    @property
    def is_optional(self) -> bool:
        return isinstance(self.type, ConceptOptionalType)

    @property
    def type(self) -> ConceptType:
        return ConceptOptionalType(inner=self._type) if not isinstance(self._type, ConceptOptionalType) and self.default_value is not None else self._type



from __future__ import annotations

import inspect
from inspect import Signature
from types import NoneType
from typing import Any, Dict, List, Optional, Tuple, Union, get_origin

from ...api import FilterExpression, NounPhrase
from ...docstring import Docstring
from ...errors import SignatureError
from ..book_procedure_descriptor import (BookProcedureDescriptor,
                                         ConnectionRequired)
from ..book_procedure_signature import BookProcedureSignature
from ..concept_descriptor import ConceptDescriptor
from ..example import Example
from ..factory.parameter_concept import ParameterConceptFactory
from ..parameter_concept_bind import ParameterConceptBind
from ..question_descriptor import QuestionDescriptor
from .concept import ConceptFactory
from .types import (extract_and_check_question_types, extract_promise_types,
                    is_none, is_union_type)


def noun_phrase_to_parameter(noun_phrase: NounPhrase) -> str:
    name = noun_phrase.modifiers.copy() if noun_phrase.modifiers else []
    name.append(noun_phrase.head)

    return "_".join(name)


def find_noun_phrase_path(target_or_object: Optional[List[NounPhrase]], noun_phrases: List[NounPhrase]) -> Optional[List[NounPhrase]]:
    if not target_or_object:
        return None

    len_target_or_object = len(target_or_object)
    len_noun_phrases = len(noun_phrases)

    for start in range(len_target_or_object - len_noun_phrases + 1):
        if target_or_object[start : start + len_noun_phrases] == noun_phrases:
            return target_or_object[: start + len_noun_phrases]

    return None


def argument_name(python_signature: Signature, pos: Optional[int]) -> Optional[str]:
    """
    Get the name of the argument at the given position in the signature.

    Args:
        python_signature: The signature to get the argument name from.
        pos: The position of the argument to get the name from.

    Returns:
        The name of the argument at the given position, or None if the position is out of bounds.
    """
    if pos is None:
        return None

    parameters = list(python_signature.parameters.values())

    if len(parameters) > pos:
        return parameters[pos].name

    return None


def has_default(python_signature: Signature, pos: Optional[int]) -> Optional[bool]:
    """
    Check if the argument at the given position in the signature has a default value.

    Args:
        python_signature: The signature to check the default value of the argument from.
        pos: The position of the argument to check the default value of.

    Returns:
        True if the argument has a default value, False otherwise. If the position is out of bounds, None is returned.
    """
    if pos is None:
        return None

    parameters = list(python_signature.parameters.values())

    if 0 <= pos < len(parameters):
        return parameters[pos].default is not inspect.Parameter.empty

    return None


class BookProcedureFactory:
    @classmethod
    def create(
        cls,
        identifier: str,
        english_signature: BookProcedureSignature,
        python_signature: Signature,
        docstring: Docstring,
        override_connection_required: Optional[ConnectionRequired],
        is_mutation: bool,
    ) -> BookProcedureDescriptor:
        # by default no procedure is filter enabled
        filter_capable = False
        filter_argument_position: Optional[int] = None

        # by default no procedure is page enabled
        page_capable = False
        offset_argument_position: Optional[int] = None
        limit_argument_position: Optional[int] = None

        # parameter concept map
        pcm: List[ParameterConceptBind] = []

        # add mapping for each part of speech
        for pos in ("object", "target"):
            english_pos: Optional[List[NounPhrase]] = getattr(english_signature, pos, None)
            if english_pos:
                python_pos_parameter = python_signature.parameters.get(pos)
                if python_pos_parameter:
                    parameter_pos_annotation = getattr(
                        python_pos_parameter.annotation,
                        "__origin__",
                        python_pos_parameter.annotation,
                    )
                    if parameter_pos_annotation and issubclass(parameter_pos_annotation, Tuple):
                        concepts = [
                            ConceptFactory.from_noun_phrase_and_annotation(
                                noun_phrase, python_pos_parameter.annotation.__args__[idx], description=docstring.input_description_by_noun_phrases([noun_phrase])
                            )
                            for idx, noun_phrase in enumerate(english_pos)
                        ]

                        pcm.append(ParameterConceptFactory.from_parameter(python_pos_parameter, concepts=concepts, description=docstring.param_description_by_name(pos)))
                    else:
                        pos_noun_phrase = NounPhrase(head=pos, modifiers=[])
                        concept = ConceptFactory.from_noun_phrase_and_annotation(pos_noun_phrase, python_pos_parameter.annotation)
                        for noun_phrases in [[pos_noun_phrase], english_pos]:
                            if noun_phrases:
                                concept.description = docstring.input_description_by_noun_phrases(noun_phrases)
                                if concept.description:
                                    break

                        pcm.append(ParameterConceptFactory.from_parameter(python_pos_parameter, concepts=[concept], description=docstring.param_description_by_name(pos)))

        # add mapping for any remaining input parameter
        for idx, p in enumerate(python_signature.parameters.items()):
            _, parameter = p

            if parameter.name == "self":
                continue

            if parameter.annotation == FilterExpression or (
                get_origin(parameter.annotation) is Union
                and len(parameter.annotation.__args__) == 2
                and parameter.annotation.__args__[0] == FilterExpression
                and parameter.annotation.__args__[1] == NoneType
            ):
                filter_argument_position = idx
                filter_capable = True
                continue

            if parameter.annotation == int or (
                get_origin(parameter.annotation) is Union
                and len(parameter.annotation.__args__) == 2
                and parameter.annotation.__args__[0] == int
                and parameter.annotation.__args__[1] == NoneType
            ):
                if parameter.name == "offset":
                    offset_argument_position = idx
                    if limit_argument_position is not None:
                        page_capable = True
                    continue

                if parameter.name == "limit":
                    limit_argument_position = idx
                    if offset_argument_position is not None:
                        page_capable = True
                    continue

            if parameter.name not in [param.python_name for param in pcm]:
                parameter_concept = ParameterConceptFactory.from_parameter(parameter, description=docstring.param_description_by_name(parameter.name))
                for concept in parameter_concept.concepts:
                    for noun_phrases in [
                        find_noun_phrase_path(english_signature.object, concept.noun_phrases),
                        find_noun_phrase_path(english_signature.target, concept.noun_phrases),
                        concept.noun_phrases,
                    ]:
                        if noun_phrases:
                            concept.description = docstring.input_description_by_noun_phrases(noun_phrases)
                            if not concept.description:
                                concept.description = parameter_concept.description

                            if concept.description:
                                break
                pcm.append(parameter_concept)

        questions, return_annotation = extract_and_check_question_types(python_signature.return_annotation)
        if questions and is_none(return_annotation) and not is_union_type(python_signature.return_annotation):
            raise ValueError(
                "A procedure cannot always/only return a question. If it can return a question, but has no results, use 'None | Question[Literal['noun phrase'], type]' as the return type-hint."
            )

        promises, return_annotation = extract_promise_types(
            return_annotation
        )  # NOTE: at this instance, we only check for promises to determine if the procedure is async (output types will be resolved later by the book decorator)

        # add type for outputs
        outputs: List[ConceptDescriptor] = []
        if english_signature and english_signature.outputs:
            if len(english_signature.outputs) > 1 and len(english_signature.outputs) != len(getattr(return_annotation, "__args__", [])) and not promises:
                raise SignatureError("The number of elements in the return tuple do not match the number of outputs in the english signature")

            for idx, output in enumerate(english_signature.outputs):
                if not promises:
                    sub_annotation: type = return_annotation.__args__[idx] if len(english_signature.outputs) > 1 else return_annotation
                else:  # NOTE: if it's an async procedure, the output type will be resolved by the book decorator (by visiting the async resolver function)
                    sub_annotation: type = Any  # type: ignore  # Any is not considered a proper type in python, but it is a valid type-hint in BDK

                outputs.append(
                    ConceptFactory.from_noun_phrases_and_annotation(
                        output,
                        sub_annotation,
                        description=docstring.output_description_by_noun_phrases(output),
                    )
                )
        elif return_annotation and return_annotation not in [None, NoneType, inspect.Signature.empty]:
            answer_noun_phrase = NounPhrase(head="answer", modifiers=[])
            output_annotation = (
                return_annotation if not promises else Any
            )  # NOTE: if it's an async procedure, the output type will be resolved by the book decorator (by visiting the async resolver function)
            outputs.append(
                ConceptFactory.from_noun_phrase_and_annotation(
                    answer_noun_phrase,
                    output_annotation,  # type: ignore  # Any is not considered a proper type in python, but it is a valid type-hint in BDK
                    description=(docstring.output_description_by_noun_phrases([answer_noun_phrase])),
                )
            )

        if len(outputs) == 1 and outputs[0].description is None and docstring.returns:
            outputs[0].description = docstring.returns

        return BookProcedureDescriptor(
            id=identifier,
            english_signature=english_signature,
            parameter_concept_map=pcm,
            outputs=outputs,
            questions=questions,
            filter_capable=filter_capable,
            filter_argument_name=argument_name(python_signature, filter_argument_position),
            filter_has_default=has_default(python_signature, filter_argument_position),
            page_capable=page_capable,
            offset_argument_name=argument_name(python_signature, offset_argument_position),
            offset_has_default=has_default(python_signature, offset_argument_position),
            limit_argument_name=argument_name(python_signature, limit_argument_position),
            limit_has_default=has_default(python_signature, limit_argument_position),
            connection_required=ConnectionRequired.OPTIONAL,
            override_connection_required=override_connection_required,
            short_description=docstring.short_description.strip() if docstring.short_description else None,
            long_description=docstring.long_description.strip() if docstring.long_description else None,
            examples=[Example.from_docstring(docstring_example) for docstring_example in docstring.examples],
            is_async=bool(promises),
            is_mutation=is_mutation,
        )

    @classmethod
    def create_discovered_procedure(
        cls,
        procedure_id: str,
        english: str,
        inputs: Dict[str, ConceptDescriptor],
        outputs: List[ConceptDescriptor],
        is_mutation: bool,
        questions: Optional[List[QuestionDescriptor]] = None,
        connection_required: ConnectionRequired = ConnectionRequired.OPTIONAL,
        filter_capable: bool = False,
        page_capable: bool = False,
    ):
        """
        Creates a BookProcedureDescriptor for a discover function.

        Args:
            procedure_id: The identifier for the procedure. It must be unique and is expected to contain all the information required by the developer to implement the invoke function.
            english: The english description of the procedure
            inputs: The input concepts of the procedure. The keys are the names of the parameters, and the values are the ConceptDescriptors.
            outputs: The output concepts of the procedure
            questions: The questions that the procedure can ask
            connection_required: Whether the procedure requires a connection
            filter_capable: Whether the procedure supports filtering. If true a `filter_expression` argument of type FilterExpression will be added.
            page_capable: Whether the procedure supports paging. If true an `offset` and `limit` argument will be added.
        """

        english_signature = BookProcedureSignature.from_english(english)

        parameter_concept_map = [ParameterConceptBind(python_name=python_name, concepts=[concept_descriptor]) for python_name, concept_descriptor in inputs.items()]

        return BookProcedureDescriptor(
            id=procedure_id,
            english_signature=english_signature,
            outputs=outputs,
            questions=questions,
            filter_capable=filter_capable,
            page_capable=page_capable,
            parameter_concept_map=parameter_concept_map,
            connection_required=connection_required,
            override_connection_required=None,
            short_description="",
            long_description="",
            examples=[],
            is_discovered=True,
            filter_argument_name=None if not filter_capable else "filter_expression",
            filter_has_default=None if not filter_capable else False,
            offset_argument_name=None if not page_capable else "offset",
            offset_has_default=None if not page_capable else False,
            limit_argument_name=None if not page_capable else "limit",
            limit_has_default=None if not page_capable else False,
            is_mutation=is_mutation,
        )



from __future__ import annotations

from dataclasses import dataclass

from ..docstring.example import DocstringExample


@dataclass
class Example:
    description: str
    snippet: str

    @classmethod
    def from_docstring(cls, docstring_example: DocstringExample) -> Example:
        return Example(description=docstring_example.description, snippet=docstring_example.snippet)



from typing import List, Optional, Set, Type

from ...api.noun_phrase import NounPhrase
from .base import ConceptType


class ConceptOpaqueType(ConceptType):
    is_a: Set[NounPhrase]
    description: Optional[str]
    concrete: Type

    def __init__(self, is_a: Set[NounPhrase], description: Optional[str], concrete: Type):
        if not isinstance(is_a, set) or not all(isinstance(item, NounPhrase) for item in is_a):
            raise TypeError("is_a must be a set of NounPhrase instances")

        self.is_a = is_a
        self.description = description
        self.concrete = concrete

    def simplify(self) -> ConceptType:
        return self

    def children(self) -> List[ConceptType]:
        return []

    def __repr__(self) -> str:
        return f"Opaque[{','.join(is_a.to_string() for is_a in list(self.is_a))}]"

    def __str__(self) -> str:
        if len(self.is_a) == 1:
            return next(iter(self.is_a)).to_string()

        return " or ".join([f"`{is_a.__str__()}`" for is_a in list(self.is_a)])

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return self

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptOpaqueType):
            return self.is_a == other.is_a
        return False

    def __hash__(self) -> int:
        combined_hash = hash("Opaque")
        for inner in self.is_a:
            combined_hash ^= hash(inner)
        return combined_hash



from ..reflection.book_procedure_descriptor import ConnectionRequired
from ..typing import Sensitive
from .discoverable import Discoverable
from .errors import (IntermediatePromiseResolverMismatchOutputError,
                     InvalidAsyncProcedureTypeError,
                     InvalidFinalPromiseResolverOutputCountError,
                     InvalidIntermediatePromiseResolverOutputCountError,
                     InvalidPromiseOutputCountError,
                     InvalidPromiseResolverParameterCountError,
                     InvalidPromiseResolverParameterTypeError,
                     InvalidPromiseTypeArgumentCountError,
                     InvalidUserInfoOutputType,
                     MissingInputPromiseInOutputError,
                     MissingPromiseResolverError, NotFoundError,
                     PromiseMismatchOutputTypeError,
                     PromiseResolverFunctionNotFoundError,
                     ResolverPromiseTypeMismatchError, TypeMismatchError,
                     UnresolvedPromiseError)
from .filter import (FilterBinaryExpression, FilterBinaryOperator,
                     FilterExpression, FilterExpressionVisitor,
                     FilterUnaryExpression, FilterUnaryOperator,
                     NounPhrasesExpression, ValueExpression)
from .noun_phrase import NounPhrase, NounPhrases
from .promise import Promise
from .questions import (Question, ask, clear_answers, get_from_context,
                        set_answer, unset_answer)
from .user_info import UserInfo

__all__ = [
    "ConnectionRequired",
    "NotFoundError",
    "TypeMismatchError",
    "FilterBinaryExpression",
    "FilterBinaryOperator",
    "FilterExpression",
    "FilterExpressionVisitor",
    "FilterUnaryExpression",
    "FilterUnaryOperator",
    "NounPhrase",
    "NounPhrases",
    "Promise",
    "Sensitive",
    "NounPhrasesExpression",
    "ValueExpression",
    "ask",
    "clear_answers",
    "get_from_context",
    "set_answer",
    "unset_answer",
    "Question",
    "Discoverable",
    "UserInfo",
    "InvalidAsyncProcedureTypeError",
    "MissingPromiseResolverError",
    "PromiseResolverFunctionNotFoundError",
    "InvalidPromiseOutputCountError",
    "PromiseMismatchOutputTypeError",
    "InvalidIntermediatePromiseResolverOutputCountError",
    "IntermediatePromiseResolverMismatchOutputError",
    "MissingInputPromiseInOutputError",
    "InvalidPromiseResolverParameterCountError",
    "InvalidPromiseResolverParameterTypeError",
    "InvalidPromiseTypeArgumentCountError",
    "ResolverPromiseTypeMismatchError",
    "UnresolvedPromiseError",
    "InvalidFinalPromiseResolverOutputCountError",
    "InvalidUserInfoOutputType",
]



import inspect
from typing import Callable, get_origin

from kognitos.bdk.api.errors import (
    IntermediatePromiseResolverMismatchOutputError,
    InvalidAsyncProcedureTypeError,
    InvalidFinalPromiseResolverOutputCountError,
    InvalidIntermediatePromiseResolverOutputCountError,
    InvalidPromiseOutputCountError, InvalidPromiseResolverParameterCountError,
    InvalidPromiseResolverParameterTypeError,
    InvalidPromiseTypeArgumentCountError, MissingInputPromiseInOutputError,
    MissingPromiseResolverError, PromiseMismatchOutputTypeError,
    PromiseResolverFunctionNotFoundError, ResolverPromiseTypeMismatchError,
    UnresolvedPromiseError)

from ...api import Promise
from ...reflection.factory.types import (ConceptTypeFactory,
                                         extract_and_check_question_types,
                                         extract_promise_types,
                                         get_inputs_annotations, is_none)


class AsyncProcedureRule:
    def validate(self, book_class: type, async_func: Callable) -> None:
        """Basic validation"""
        if not inspect.isfunction(async_func):
            raise InvalidAsyncProcedureTypeError("The async procedure rule can only be validated against functions.")

        if not hasattr(async_func, "__promise_resolver_function_name__"):
            raise MissingPromiseResolverError(f"Missing promise resolver for function '{async_func.__name__}'. Did you forget to apply the @promise decorator?")

        if not hasattr(book_class, async_func.__promise_resolver_function_name__):
            raise PromiseResolverFunctionNotFoundError(
                f"Promise resolver '{async_func.__promise_resolver_function_name__}' for function '{async_func.__name__}' not found in book class."
            )


class AsyncProcedureOutputsRule(AsyncProcedureRule):
    def validate(self, book_class: type, async_func: Callable) -> None:
        super().validate(book_class, async_func)
        _, async_func_output = extract_and_check_question_types(inspect.signature(async_func).return_annotation)
        async_func_promises, async_func_return_annotation = extract_promise_types(async_func_output)
        if getattr(async_func, "__is_async__", False) and hasattr(async_func, "__procedure__"):
            if len(async_func_promises) != 1:
                raise InvalidPromiseOutputCountError(
                    f"Promise procedure <{async_func.__name__}> must output exactly one promise. Got {len(async_func_promises)} in the type-hint instead."
                )
            if not is_none(async_func_return_annotation):
                raise PromiseMismatchOutputTypeError(
                    f"Promise procedure <{async_func.__name__}> should not output any other types when returning a promise. Got {async_func_return_annotation}."
                )


class IntermediateAsyncResolverOutputsRule(AsyncProcedureRule):
    def validate(self, book_class: type, async_func: Callable) -> None:
        intermediate_async_resolver = async_func
        super().validate(book_class, intermediate_async_resolver)
        _, intermediate_async_resolver_output = extract_and_check_question_types(inspect.signature(intermediate_async_resolver).return_annotation)
        intermediate_async_resolver_promises, intermediate_async_resolver_return_annotation = extract_promise_types(intermediate_async_resolver_output)
        if getattr(intermediate_async_resolver, "__is_async__", False) and not hasattr(intermediate_async_resolver, "__procedure__"):
            # NOTE: at this point, its guaranteed that intermediate_async_resolver has a promise resolver itself (we are in an intermediate resolver)
            if len(intermediate_async_resolver_promises) not in [1, 2]:
                raise InvalidIntermediatePromiseResolverOutputCountError(
                    f"Intermediate promise resolver <{intermediate_async_resolver.__name__}> must be able to output the input and output promises (two if they are of different types, one if they are the same). Got {len(intermediate_async_resolver_promises)} promise declarations in the type-hint instead."
                )
            if not is_none(intermediate_async_resolver_return_annotation):
                raise IntermediatePromiseResolverMismatchOutputError(
                    f"Intermediate promise resolver <{intermediate_async_resolver.__name__}> should not output anything other than the input/output promises. Got {intermediate_async_resolver_return_annotation}."
                )

            # If we reach this point, we know for a fact that intermediate_async_resolver has only one input, which is of type Promise (matching the output of the previous function)
            # Check if it can also return the promise it received
            intermediate_async_resolver_input = get_inputs_annotations(intermediate_async_resolver)
            intermediate_async_resolver_input_annotation = next(iter(intermediate_async_resolver_input.values()))
            intermediate_async_resolver_input_type = ConceptTypeFactory.from_type(intermediate_async_resolver_input_annotation.__args__[0])
            if intermediate_async_resolver_input_type not in intermediate_async_resolver_promises:
                raise MissingInputPromiseInOutputError(
                    f"Intermediate promise resolver <{intermediate_async_resolver.__name__}> must be able to return the promise it received as input. Takes [{repr(intermediate_async_resolver_input_type)}], can return {repr(intermediate_async_resolver_promises)}."
                )


class AsyncResolverInputsRule(AsyncProcedureRule):
    def validate(self, book_class: type, async_func: Callable) -> None:
        super().validate(book_class, async_func)
        _, async_func_output = extract_and_check_question_types(inspect.signature(async_func).return_annotation)
        async_func_promises, _ = extract_promise_types(async_func_output)
        resolver_func = getattr(book_class, async_func.__promise_resolver_function_name__)  # type: ignore # We already checked for the existence of the resolver in the rule superclass
        resolver_input = get_inputs_annotations(resolver_func)

        if len(resolver_input) != 1:
            raise InvalidPromiseResolverParameterCountError(
                f"Promise resolver '{async_func.__promise_resolver_function_name__}' for function '{async_func.__name__}' must have exactly one input parameter. Got {len(resolver_input)} instead."  # type: ignore # We already checked for the existence of the resolver in the rule superclass
            )
        resolver_input_annotation = next(iter(resolver_input.values()))
        if get_origin(resolver_input_annotation) is not Promise:
            raise InvalidPromiseResolverParameterTypeError(
                f"Promise resolver '{async_func.__promise_resolver_function_name__}' for function '{async_func.__name__}' must take a promise as its input parameter. Got {resolver_input_annotation} instead."  # type: ignore # We already checked for the existence of the resolver in the rule superclass
            )
        resolver_input_type = ConceptTypeFactory.from_type(resolver_input_annotation.__args__[0])
        if len(resolver_input_annotation.__args__) != 1:
            raise InvalidPromiseTypeArgumentCountError("Promise must have exactly one type argument")

        if (
            len(async_func_promises) > 1
        ):  # If the async function outputs more than one promise, we discard the one it receives as input (if it outputs just one promise, the input is the same type as the output)
            async_func_input = get_inputs_annotations(async_func)
            async_func_input_annotation = next(iter(async_func_input.values()))
            async_func_input_type = ConceptTypeFactory.from_type(async_func_input_annotation.__args__[0])
            async_func_promises.remove(async_func_input_type)
        if resolver_input_type not in async_func_promises:
            raise ResolverPromiseTypeMismatchError(
                f"Promise resolver <{async_func.__promise_resolver_function_name__}> must take the same promise type as one the output of the function it resolves <{async_func.__name__}>. Got [{repr(resolver_input_type)}], expected {async_func_promises}."  # type: ignore # We already checked for the existence of the resolver in the rule superclass
            )


class AsyncResolverNoUnresolvedPromisesRule(AsyncProcedureRule):
    def validate(self, book_class: type, async_func: Callable) -> None:
        super().validate(book_class, async_func)
        _, async_func_output = extract_and_check_question_types(inspect.signature(async_func).return_annotation)
        async_func_promises, _ = extract_promise_types(async_func_output)
        unresolved_promises = set(async_func_promises)

        # Discard the <input promise> output (in case of an intermediate resolver)
        if not hasattr(async_func, "__procedure__"):
            async_func_input = get_inputs_annotations(async_func)
            async_func_input_annotation = next(iter(async_func_input.values()))
            async_func_input_type = ConceptTypeFactory.from_type(async_func_input_annotation.__args__[0])
            unresolved_promises.discard(async_func_input_type)

        if unresolved_promises:  # If there are still promises to resolve, check the resolver
            resolver_func = getattr(book_class, async_func.__promise_resolver_function_name__)  # type: ignore # We already checked for the existence of the resolver in the rule superclass
            resolver_input = get_inputs_annotations(resolver_func)
            resolver_input_annotation = next(iter(resolver_input.values()))  # We know for a fact that any resolver has exactly one input (a promise)
            resolver_input_type = ConceptTypeFactory.from_type(resolver_input_annotation.__args__[0])
            unresolved_promises.discard(resolver_input_type)

        if unresolved_promises:
            raise UnresolvedPromiseError(
                f"Promise function <{async_func.__name__}> outputs unresolved promises (neither resolved by the function itself nor by its resolver): {unresolved_promises}"
            )


class AsyncFinalResolverOutputsRule(AsyncProcedureRule):
    def validate(self, book_class: type, async_func: Callable) -> None:
        super().validate(book_class, async_func)
        resolver_func = getattr(book_class, async_func.__promise_resolver_function_name__)  # type: ignore # We already checked for the existence of the resolver in the rule superclass
        if not getattr(resolver_func, "__is_async__", False):  # If it's not async, it must be the final resolver
            _, resolver_output = extract_and_check_question_types(inspect.signature(resolver_func).return_annotation)
            resolver_promises, _ = extract_promise_types(resolver_output)
            if len(resolver_promises) != 1:
                raise InvalidFinalPromiseResolverOutputCountError(
                    f"Final promise resolver <{async_func.__promise_resolver_function_name__}> should be able to output one promise (the one it received as input). Got {len(resolver_promises)} in the type-hint instead."  # type: ignore # We already checked for the existence of the resolver in the rule superclass
                )



from inspect import Parameter
from typing import Any, List, Optional

from ...api.noun_phrase import NounPhrase
from ...reflection import ConceptDescriptor
from .types import ConceptTypeFactory


class ConceptFactory:
    @classmethod
    def from_parameter(cls, parameter: Parameter) -> "ConceptDescriptor":
        noun_phrase = NounPhrase.from_word_list(parameter.name.split("_"))

        default_value = parameter.default if parameter.default is not Parameter.empty else None

        return ConceptFactory.from_noun_phrase_and_annotation(noun_phrase, parameter.annotation, default_value=default_value)

    @classmethod
    def from_noun_phrase_and_annotation(
        cls, noun_phrase: NounPhrase, annotation: "type", description: Optional[str] = None, default_value: Optional[Any] = None
    ) -> "ConceptDescriptor":
        return ConceptFactory.from_noun_phrases_and_annotation([noun_phrase], annotation, description, default_value=default_value)

    @classmethod
    def from_noun_phrases_and_annotation(
        cls, noun_phrases: List[NounPhrase], annotation: "type", description: Optional[str] = None, default_value: Optional[Any] = None
    ) -> "ConceptDescriptor":
        t = ConceptTypeFactory.from_type(annotation)

        return ConceptDescriptor(noun_phrases=noun_phrases, type=t, description=description, default_value=default_value)



import functools
import inspect

from ..docstring import DocstringParser
from ..reflection import CredentialDescriptor
from ..reflection.factory import BookCustomAuthenticationFactory
from ..reflection.types import CredentialScalarType
from ..typing import Sensitive


def connect(noun_phrase: str, name: str):
    """
    Function decorator for registering a custom connection method in a book.

    Args:
        noun_phrase (str): The noun phrase that identifies the connection method (e.g., "api key method").
        name (str): A display name for the connection method (e.g., "Api Key").
    """

    def decorator(fn):
        if not inspect.isfunction(fn):
            raise TypeError("The connect decorator can only be applied to functions.")

        short_description = None
        signature = inspect.signature(fn)
        credentials = []

        if not fn.__doc__:
            raise ValueError("missing docstring")

        parsed_docstring = DocstringParser.parse(fn.__doc__)

        if parsed_docstring.short_description:
            short_description = parsed_docstring.short_description.strip()

        for param_name, parameter in signature.parameters.items():
            if param_name == "self":
                continue

            # Handle special 'verify' parameter
            if param_name == "verify":
                # Validate that verify is a boolean
                if parameter.annotation is not bool:
                    raise TypeError("parameter 'verify' must be a boolean and must be annotated as bool")
                # Skip adding verify as a credential
                continue

            if parameter.annotation is str:
                credential_type = CredentialScalarType.TEXT
            elif parameter.annotation is Sensitive[str]:
                credential_type = CredentialScalarType.SENSITIVE
            else:
                # NOTE: We don't support other types for now
                raise TypeError(f"parameter '{param_name}' has an unsupported type")

            if parsed_docstring:
                param_docstring = parsed_docstring.param_by_name(param_name)
            else:
                param_docstring = None

            credential = CredentialDescriptor(
                id=param_name,
                label=param_docstring.label if param_docstring is not None else None,
                type=credential_type,
                description=(param_docstring.description if param_docstring is not None else None),
            )

            credentials.append(credential)

        if not hasattr(fn, "__connect__"):
            fn.__connect__ = BookCustomAuthenticationFactory.create(id=fn.__name__, noun_phrase_str=noun_phrase, credentials=credentials, description=short_description, name=name)

        if not hasattr(fn, "__signature__"):
            fn.__signature__ = signature

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    return decorator



from typing import Literal, get_origin

from ...api import Question
from ...api.noun_phrase import NounPhrase, NounPhrases
from ...reflection.factory.types import ConceptTypeFactory, is_union_type
from ...reflection.types.opaque import ConceptOpaqueType


class QuestionTypingRule:
    def validate(self, annotation_type: type) -> None:
        if annotation_type is Question:
            raise ValueError("Question type-hinting must be parameterized: Question[Literal['noun phrase'], type]")


class QuestionTypeHintStructureRule(QuestionTypingRule):
    def validate(self, annotation_type: type) -> None:
        super().validate(annotation_type)
        if (
            not len(annotation_type.__args__) == 2
            or not get_origin(annotation_type.__args__[0]) is Literal
            or not len(annotation_type.__args__[0].__args__) == 1
            or not isinstance(annotation_type.__args__[0].__args__[0], str)
            or not (isinstance(annotation_type.__args__[1], type) or is_union_type(annotation_type.__args__[1]))
        ):
            raise ValueError("Question type-hinting must have a string literal as the first argument, and a type as the second argument.")


class QuestionTypeHintNounPhraseRule(QuestionTypingRule):
    def validate(self, annotation_type: type) -> None:
        super().validate(annotation_type)
        noun_phrases = NounPhrases([NounPhrase.from_str(noun_phrase) for noun_phrase in annotation_type.__args__[0].__args__[0].split("'s ")])
        if not noun_phrases.noun_phrases:
            raise ValueError("Question type-hinting must have a valid noun phrases representation in the string literal.")


class QuestionTypeHintConceptTypeRule(QuestionTypingRule):
    def validate(self, annotation_type: type) -> None:
        super().validate(annotation_type)
        question_concept_type = ConceptTypeFactory.from_type(annotation_type.__args__[1])
        if question_concept_type == ConceptOpaqueType({NounPhrase("thing")}, "", object):
            raise ValueError("Question type-hinting must declare a supported concept type.")



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

from .book_procedure_signature import BookProcedureSignature
from .concept_descriptor import ConceptDescriptor
from .example import Example
from .parameter_concept_bind import ParameterConceptBind
from .question_descriptor import QuestionDescriptor


class ConnectionRequired(Enum):
    OPTIONAL = 0
    ALWAYS = 1
    NEVER = 2


@dataclass
class BookProcedureDescriptor:
    id: str
    english_signature: BookProcedureSignature
    examples: List[Example] = field(default_factory=list)
    filter_capable: bool = False
    page_capable: bool = False
    connection_required: ConnectionRequired = ConnectionRequired.OPTIONAL
    is_discovered: bool = False
    override_connection_required: Optional[ConnectionRequired] = None
    outputs: Optional[List[ConceptDescriptor]] = None
    questions: Optional[List[QuestionDescriptor]] = None
    short_description: Optional[str] = None
    long_description: Optional[str] = None
    parameter_concept_map: List[ParameterConceptBind] = field(default_factory=list)
    filter_argument_name: Optional[str] = None
    offset_argument_name: Optional[str] = None
    limit_argument_name: Optional[str] = None
    filter_has_default: Optional[bool] = None
    offset_has_default: Optional[bool] = None
    limit_has_default: Optional[bool] = None
    is_async: bool = False
    is_mutation: bool = True

    @property
    def input_concepts(self) -> Optional[List[ConceptDescriptor]]:
        input_concepts = []

        for parameter_concept in self.parameter_concept_map:
            input_concepts.extend(parameter_concept.concepts)

        return input_concepts if len(input_concepts) != 0 else None

    @property
    def output_concepts(self) -> Optional[List[ConceptDescriptor]]:
        return self.outputs



from dataclasses import dataclass
from typing import Optional


@dataclass
class Discoverable:
    """
    Represents a discoverable object.
    This is to be returned by the @discoverables method of a book.

    The idea is to provide the user with knowledge about the different discoverable
    entities in the system for them to properly call the discover mechanism.

    Args:
        name: The name of the discoverable.
        description: The description of the discoverable.
    """

    name: str
    description: Optional[str] = None



from typing import Any, List, Optional, Set, Type

from ...api.noun_phrase import NounPhrase
from .base import ConceptType
from .optional import ConceptOptionalType


class ConceptDictionaryTypeField:
    name: str
    description: Optional[str]
    _t: ConceptType
    default_value: Any
    has_default_value: bool
    init: bool

    def __init__(
        self, name: str, description: Optional[str] = None, t: Optional[ConceptType] = None, default_value: Any = None, has_default_value: bool = False, init: bool = True
    ):
        self.name = name
        self.description = description
        self._t = t
        self.default_value = default_value
        self.has_default_value = has_default_value
        self.init = init

    def __repr__(self) -> str:
        return f"{self.name}: {self.t!r}"

    def __str__(self) -> str:
        return f"{self.name}: {self.t}"

    def __eq__(self, other) -> bool:
        if self.name != other.name:
            return False
        if self.t != other.t:
            return False
        if self.description != other.description:
            return False
        return True

    def __hash__(self) -> int:
        return hash((self.name, self.description, self.t))

    @property
    def t(self) -> ConceptType:
        return ConceptOptionalType(inner=self._t) if not isinstance(self._t, ConceptOptionalType) and self.default_value is not None else self._t


class ConceptDictionaryType(ConceptType):
    is_a: Set[NounPhrase]
    description: Optional[str]
    concrete: Optional[Type]
    unset: Optional[Any]

    def __init__(self, is_a: Set[NounPhrase], concrete: Optional[Type], description: Optional[str], fields: List[ConceptDictionaryTypeField], unset: Optional[Any] = None):
        self.is_a = is_a
        self.description = description
        self.fields = fields
        self.concrete = concrete
        self.unset = unset

    def simplify(self) -> ConceptType:
        simplified_fields = [
            ConceptDictionaryTypeField(field.name, field.description, field.t.simplify(), field.default_value, field.has_default_value, field.init) for field in self.fields
        ]

        return ConceptDictionaryType(self.is_a, self.concrete, self.description, simplified_fields, self.unset)

    def children(self) -> List[ConceptType]:
        return [field.t for field in self.fields]

    def __repr__(self) -> str:
        return f"Dictionary[({','.join([is_a.to_string() for is_a in list(self.is_a)])}), {', '.join([field.__repr__() for field in self.fields])}]"

    def __str__(self) -> str:
        if len(self.is_a) == 1:
            return next(iter(self.is_a)).to_string()

        return " or ".join([f"`{is_a.__str__()}`" for is_a in list(self.is_a)])

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new

        new_fields = [
            ConceptDictionaryTypeField(field.name, field.description, new) if field.t == old else ConceptDictionaryTypeField(field.name, field.description, old)
            for field in self.fields
        ]

        return ConceptDictionaryType(self.is_a, self.concrete, self.description, new_fields, self.unset)

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptDictionaryType):
            return self.concrete == other.concrete and self.is_a == other.is_a and self.fields == other.fields and self.unset == other.unset
        return False

    def __hash__(self) -> int:
        combined_hash = hash("Dictionary")
        for field in self.fields:
            combined_hash ^= hash(field)
        return combined_hash



import threading
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union

from ..reflection.types.base import ConceptType
from .noun_phrase import NounPhrases

T = TypeVar("T")
NounPhrasesStringLiteral = TypeVar("NounPhrasesStringLiteral", bound=str)


class Question(Generic[NounPhrasesStringLiteral, T]):
    """
    Represents a question to be asked to the user.

    Args:
        concept_name: The name of the concept whose value you want to ask for.
        concept_type: The type that you expect the answer to be.
        choices: The choices for the user to select (if any).
        text: The text you want to present to the user (if any).
    """

    noun_phrases: NounPhrases
    concept_type: ConceptType
    choices: List[T]
    text: Optional[str]

    def __init__(self, concept_name: NounPhrasesStringLiteral, concept_type: Type[T], choices: Optional[List[T]] = None, text: Optional[str] = None):
        """
        Initialize a Question object.

        Args:
            concept_name: The name of the concept whose value you want to ask for.
            concept_type: The type that you expect the answer to be.
            choices: The choices for the user to select (if any).
            text: The text you want to present to the user (if any).
        """
        from ..reflection.factory.types import ConceptTypeFactory

        self.noun_phrases = NounPhrases.from_str(concept_name)
        self.concept_type = ConceptTypeFactory.from_type(concept_type)
        self.choices = choices if choices else []
        self.text = text

    def __hash__(self) -> int:
        return hash(self.noun_phrases)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Question) and hash(self) == hash(other)

    def __str__(self) -> str:
        ret = "Question("
        nps_text = str(self.noun_phrases)
        if self.text:
            ret += f"{self.text}, "

        ret += f"noun_phrases='{nps_text}'"

        if self.choices:
            ret += f", choices={self.choices}"

        return ret + ")"

    def __repr__(self) -> str:
        return str(self)


class AnswerStorage(ABC):
    @abstractmethod
    def get(self, concept_noun_phrases: NounPhrases) -> Optional[Any]:
        raise NotImplementedError()

    @abstractmethod
    def set(self, concept_noun_phrases: NounPhrases, answer_value: Any):
        raise NotImplementedError()

    @abstractmethod
    def clear(self):
        raise NotImplementedError()

    @abstractmethod
    def unset(self, concept_noun_phrases: NounPhrases, fail_if_not_set: bool):
        raise NotImplementedError()


class InThreadMemoryAnswerStorage(AnswerStorage):
    def __init__(self):
        self._answer_map: Dict[int, Dict[str, Any]] = {}

    @property
    def thread_id(self) -> int:
        return threading.get_ident()

    def get(self, concept_noun_phrases: NounPhrases) -> Optional[Any]:
        concept_name = str(concept_noun_phrases)
        return self._answer_map.get(self.thread_id, {}).get(concept_name)

    def set(self, concept_noun_phrases: NounPhrases, answer_value: Any):
        concept_name = str(concept_noun_phrases)
        if self.thread_id not in self._answer_map:
            self._answer_map[self.thread_id] = {}
        self._answer_map[self.thread_id][concept_name] = answer_value

    def unset(self, concept_noun_phrases: NounPhrases, fail_if_not_set: bool):
        concept_name = str(concept_noun_phrases)
        if self.thread_id in self._answer_map and concept_name in self._answer_map:
            del self._answer_map[self.thread_id][concept_name]
        elif fail_if_not_set:
            raise ValueError(f"Key '{concept_name}' is not set")

    def clear(self):
        if self.thread_id in self._answer_map:
            del self._answer_map[self.thread_id]


# Use the InThreadMemoryAnswerStorage by default
answer_storage = InThreadMemoryAnswerStorage()


def get_from_context(concept_noun_phrases: NounPhrases) -> Optional[Any]:
    return answer_storage.get(concept_noun_phrases)


def set_answer(concept_noun_phrases: NounPhrases, answer_value: Any):
    answer_storage.set(concept_noun_phrases, answer_value)


def unset_answer(concept_noun_phrases: NounPhrases, fail_if_not_set: bool = False):
    answer_storage.unset(concept_noun_phrases, fail_if_not_set)


def clear_answers():
    answer_storage.clear()


def ask(
    concept_name: NounPhrasesStringLiteral, concept_type: Type[T], text: Optional[str] = None, choices: Optional[List[T]] = None
) -> Union[T, Question[NounPhrasesStringLiteral, T]]:
    """
    Ask a question and get an answer (if available) or a Question object to be
    returned to the user.

    Args:
        concept_name: The unique name of the concept whose value you want to
            ask for.
        concept_type: The type that you expect the answer to be.
        choices: The choices for the user to select (if any).
        text: The text you want to present to the user (if any).

    Returns:
        Answer containing the value from the answer map, typed as T (the
        `concept_type`) or a Question (typed as `Question[Literal[concept_name], T]`)
        if the question is not answered yet.
    """
    possible_answer = get_from_context(NounPhrases.from_str(concept_name))
    if possible_answer is not None:
        if not choices or _value_in_choices(possible_answer, choices):
            return possible_answer

    return Question(concept_name, concept_type, choices, text)


def _value_in_choices(value: Any, choices: List[Any]) -> bool:
    if isinstance(value, Decimal):
        return any(float(choice) == float(value) for choice in choices if isinstance(choice, (Decimal, float, int)))
    return value in choices



import inspect


def promise(resolver: str):
    def decorator(fn):
        if not inspect.isfunction(fn):
            raise TypeError("The promise decorator can only be applied to functions.")

        fn.__promise_resolver_function_name__ = resolver
        fn.__is_async__ = True

        return fn

    return decorator



import inspect
from functools import wraps


def discoverables(fn):
    if not inspect.isfunction(fn):
        raise TypeError("The discoverables decorator can only be applied to functions.")

    fn.__discoverables__ = True
    return wraps(fn)(fn)



from typing import List

from .base import ConceptType
from .self import ConceptSelfType


class ConceptListType(ConceptType):
    def __init__(self, inner: ConceptType):
        self.type = inner

    def simplify(self) -> ConceptType:
        simplify_inner = self.type.simplify()

        if isinstance(simplify_inner, ConceptListType) and isinstance(simplify_inner.type, ConceptSelfType):
            return simplify_inner

        return ConceptListType(simplify_inner)

    def children(self) -> List[ConceptType]:
        return [self.type]

    def __repr__(self) -> str:
        return f"List[{self.type.__repr__()}]"

    def __str__(self) -> str:
        return f"list of {self.type.name}"

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return ConceptListType(self.type.replace(old, new))

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptListType):
            return self.type == other.type
        return False

    def __hash__(self) -> int:
        combined_hash = hash("List")
        combined_hash ^= hash(self.type)
        return combined_hash



from .promise_rules import (AsyncFinalResolverOutputsRule,
                            AsyncProcedureOutputsRule, AsyncProcedureRule,
                            AsyncResolverInputsRule,
                            AsyncResolverNoUnresolvedPromisesRule,
                            IntermediateAsyncResolverOutputsRule)
from .question_rules import (QuestionTypeHintConceptTypeRule,
                             QuestionTypeHintNounPhraseRule,
                             QuestionTypeHintStructureRule, QuestionTypingRule)

__all__ = [
    "AsyncFinalResolverOutputsRule",
    "AsyncProcedureRule",
    "AsyncProcedureOutputsRule",
    "AsyncResolverInputsRule",
    "AsyncResolverNoUnresolvedPromisesRule",
    "IntermediateAsyncResolverOutputsRule",
    "QuestionTypeHintConceptTypeRule",
    "QuestionTypeHintNounPhraseRule",
    "QuestionTypeHintStructureRule",
    "QuestionTypingRule",
]



import inspect
from functools import wraps

from kognitos.bdk.api.errors import InvalidUserInfoOutputType
from kognitos.bdk.api.user_info import UserInfo


def userinfo(fn):
    """
    Function decorator for flagging the user information function in a book.

    A user information function outputs information about the authenticated user in an OAuth connected book.
    """

    if not inspect.isfunction(fn):
        raise TypeError("The userinfo decorator can only be applied to functions.")

    output = inspect.signature(fn).return_annotation
    if output is not UserInfo:
        raise InvalidUserInfoOutputType(f"Function <{fn.__name__}> decorated with @userinfo must output a UserInfo object")

    fn.__userinfo__ = True

    return wraps(fn)(fn)



import inspect
from typing import Optional

from kognitos.bdk.api import NounPhrase
from kognitos.bdk.klang import KlangParser


def config(*args, **kwargs):
    name: Optional[str] = None

    if len(args) == 1 and inspect.isfunction(args[0]):
        fn = args[0]
        name = None
        default_value = None
    elif kwargs:
        fn = None
        name = kwargs.get("name", None)
        if not name and len(args) == 1:
            name = args[0]
        default_value = kwargs.get("default_value", None)
    else:
        fn = None
        name = args[0]
        default_value = None

    def decorator(fn):
        if not inspect.isfunction(fn):
            raise TypeError("The config decorator can only be applied to functions which also have @property.")

        # parse noun phrases
        if name:
            noun_phrases_tuples, _ = KlangParser.parse_noun_phrases(name)
            noun_phrases = [NounPhrase.from_tuple(npt) for npt in noun_phrases_tuples]
        else:
            noun_phrases = [NounPhrase.from_word_list(fn.__name__.split("_"))]

        if not hasattr(fn, "__noun_phrase__"):
            fn.__noun_phrase__ = noun_phrases

        if not hasattr(fn, "__default_value__"):
            fn.__default_value__ = default_value

        return fn

    if fn is not None:
        return decorator(fn)

    return decorator



from typing import List

from .base import ConceptType
from .optional import ConceptOptionalType, contains_optionals


class ConceptSensitiveType(ConceptType):
    def __init__(self, inner: ConceptType):
        self.type = inner

    def simplify(self) -> ConceptType:
        flattened_inners = self.flatten()
        simplified = flattened_inners.pop()  # NOTE: flattening a sensitive should always yield a list of one element
        if contains_optionals(self):
            return ConceptOptionalType(simplified)
        return simplified

    def children(self) -> List[ConceptType]:
        return [self.type]

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return ConceptSensitiveType(self.type.replace(old, new))

    def flatten(self, **kwargs) -> List[ConceptType]:
        from .union import ConceptUnionType  # pylint: disable=cyclic-import

        strip_sensitives = kwargs.get("strip_sensitive", False)

        kwargs.update(strip_sensitive=True)
        flattened_types = self.type.flatten(**kwargs)

        if strip_sensitives is True:
            return flattened_types

        if len(flattened_types) == 1:
            return [ConceptSensitiveType(flattened_types.pop())]

        return [ConceptSensitiveType(ConceptUnionType(flattened_types))]

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptSensitiveType):
            return self.type == other.type
        return False

    def __repr__(self) -> str:
        return f"Sensitive[{self.type.name}]"

    def __str__(self) -> str:
        return f"Sensitive[{self.type.name}]"

    def __hash__(self) -> int:
        combined_hash = hash("Sensitive")
        combined_hash ^= hash(self.type)
        return combined_hash



from .authentication import (BookAuthenticationDescriptor,
                             BookCustomAuthenticationDescriptor,
                             BookOAuthAuthenticationDescriptor,
                             CredentialDescriptor, OauthArgumentDescriptor,
                             OAuthFlow, OAuthProvider, OAuthTokenFunction)
from .book_descriptor import BookDescriptor
from .book_procedure_descriptor import (BookProcedureDescriptor,
                                        ConnectionRequired)
from .book_procedure_signature import BookProcedureSignature
from .concept_descriptor import ConceptDescriptor
from .parameter_concept_bind import ParameterConceptBind
from .question_descriptor import QuestionDescriptor
from .types import (ConceptAnyType, ConceptDictionaryType,
                    ConceptDictionaryTypeField, ConceptEnumType,
                    ConceptListType, ConceptOpaqueType, ConceptOptionalType,
                    ConceptScalarType, ConceptSelfType, ConceptSensitiveType,
                    ConceptTableType, ConceptType, ConceptUnionType,
                    CredentialScalarType, CredentialType)



from typing import List

from .base import ConceptType


class ConceptAnyType(ConceptType):
    def simplify(self) -> ConceptType:
        return self

    def children(self) -> List[ConceptType]:
        return []

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new
        return self

    def __repr__(self):
        return "Any"

    def __str__(self):
        return "any"

    def __eq__(self, other):
        if isinstance(other, ConceptAnyType):
            return True
        return False

    def __hash__(self):
        return hash("Any")



from typing import Optional

from ..reflection.types import ConceptType


class TypeMismatchError(Exception):
    where: str
    expected: Optional[ConceptType]

    def __init__(self, where: str, expected: ConceptType):
        self.where = where
        self.expected = expected
        super().__init__(f"type mismatch on {where} expected {expected}")


class NotFoundError(Exception):
    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class InvalidAsyncProcedureTypeError(TypeError):
    pass


class MissingPromiseResolverError(ValueError):
    pass


class PromiseResolverFunctionNotFoundError(AttributeError):
    pass


class InvalidPromiseOutputCountError(ValueError):
    pass


class PromiseMismatchOutputTypeError(TypeError):
    pass


class InvalidIntermediatePromiseResolverOutputCountError(ValueError):
    pass


class IntermediatePromiseResolverMismatchOutputError(TypeError):
    pass


class MissingInputPromiseInOutputError(TypeError):
    pass


class InvalidPromiseResolverParameterCountError(ValueError):
    pass


class InvalidPromiseResolverParameterTypeError(TypeError):
    pass


class InvalidPromiseTypeArgumentCountError(TypeError):
    pass


class ResolverPromiseTypeMismatchError(TypeError):
    pass


class UnresolvedPromiseError(ValueError):
    pass


class InvalidFinalPromiseResolverOutputCountError(ValueError):
    pass


class InvalidUserInfoOutputType(TypeError):
    pass



from typing import Generic, TypeVar

T = TypeVar("T")


class Promise(Generic[T]):
    """
    A BDK promise returned by a procedure, to be resolved by a
    resolver function (specified by the `@promise` decorator on
    the procedure itself).
    """

    def __init__(self, data: T):
        self.data = data
        self._resolver_id = None  # This should never be set manually, but instead by the BDK runtime

    def __str__(self):
        return f"Promise({self.data})"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash((self.data, self._resolver_id))

    def __eq__(self, other):
        return isinstance(other, Promise) and hash(self) == hash(other)



from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class UserInfo:
    email: Optional[str] = None
    username: Optional[str] = None
    other_attributes: Optional[Dict[str, str]] = None
